
/*
 * AUTHOR:
 *   Bernd Helmle, Jaime Casanova
 * DESCRIPTION:
 *   Functions to translate a SQL92 compliant view definition
 *   into update rules.
 * TODO:
 *   - Update from array fields (array[index]) isn't supported right now.
 */

#include <rewrite/view_update.h>
#include <nodes/makefuncs.h>
#include <rewrite/rewriteManip.h>
#include <rewrite/rewriteHandler.h>
#include <utils/lsyscache.h>
#include <optimizer/clauses.h>

struct ViewExprContext {

	Index       newRTE;
	Index       oldRTE;
	Index       baseRTE;
	Index       subQueryLevel;
	ViewDefColumnList tentries;

};

struct ViewQualCheckFunc 
{
  
	HeapTuple    heap;
	Form_pg_proc tuple;
	Oid          procoid;

};

static const char *const PG_VIEW_UPDATE_CONTROL 
= "pg_view_update_error";
static const int16 PG_VIEW_UPDATE_CONTROL_NARGS = 1;

static char * retrieve_rule(Oid,
							const char *,
							const bool noLock);

static RuleStmt *
create_nothing_rule(Query          *action,
					const RangeVar *var,
					const char     *rulename,
					CmdType        cmdType,
					bool           replace);

static RuleStmt *
create_insert_rule(Query          *insert,
				   const RangeVar *view,
				   bool           replace);

static RuleStmt *
create_update_rule(Query          *update,
				   const RangeVar *view,
				   bool           replace);

static RuleStmt *
create_delete_rule(Query          *delete,
				   const RangeVar *view,
				   bool           replace);

static Oid
hasRule(Oid        view,
		const char *name);

static void form_query(const Query *select,
					   Query       *query,
					   ViewDefColumnList tentries,
					   bool        copyTargetList);

static RangeTblEntry *
get_relation_RTE(const Query  *select,
				 unsigned int *offset);

static Index
get_rtindex_for_rel(List       *rte_list,
					const char *relname);

static struct ViewQualCheckFunc
get_view_qualification_func(const char *name);

static bool
view_def_qual_walker(Node *node,
					 struct ViewExprContext *context);

static FuncExpr *
create_qual_for_rule(RuleStmt    *stmt,
					 const Query *select,
					 ViewDefColumnList tentries,
					 Index       newRTE,
					 Index       oldRTE);

static int
replace_varnos_for_target(List        *targetList,
						  ViewDefColumnList tentries,
						  Index       rtIndex);

static bool
replace_varnos_for_var(Node *node,
					   Index rtIndex,
					   Index oldIndex, 
					   Index subQueryLevel);

static OpExpr *
create_opexpr(Var *var_left,
			  Var *var_right);

static void
form_where_for_updrule(const Query    *select,/* View retrieve rule */
					   FromExpr       **from, /* FromExpr for stmt */
					   const Relation rel,    /* base relation of view */
					   Oid            baserel,/* Oid of base relation */
					   Index          baserti,/* Index of base relation RTE */
					   Index          oldrti);/* Index of *OLD* RTE */

static Query *
transform_to_QueryTree(StringInfoData buf);

static void
build_update_target_list(const Query    *update,
						 const Query    *select,
						 ViewDefColumnList tentries,
						 Oid               baserel,
						 const Relation    rel );

static void
form_vars_arrayref(ArrayRef *array,
				   Index newRTE,
				   ViewDefColumnList tentries);

static bool
isInsertable(const Query *select, 
			 const Relation rel);

extern void
get_base_relations(struct ViewBaseRelation *tree,
				   List   **baserelations);

/*
 * System function to check CHECK OPTION expression
 */
Datum
pg_view_update_error(PG_FUNCTION_ARGS);

/*------------------------------------------------------------------------------
 * Private functions...
 * -----------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Returns the range table index for the specified relname.
 *------------------------------------------------------------------------------
 */
static Index
get_rtindex_for_rel( List       *rte_list,
                     const char *relname ) 
{

	Index result = 0;

	Assert(relname != NULL);

	if (relname != NULL) 
	{

		RangeTblEntry *rte  = NULL;
		ListCell      *cell = NULL;
		int           index = 0;

		foreach(cell, rte_list) 
		{

			rte = (RangeTblEntry *) lfirst(cell);
			++index;

			if (rte != NULL) 
			{

				if (strncmp(rte->eref->aliasname, relname, strlen(relname)) == 0) 
				{

					result = index;
					break;

				}

			}

		}

	} else
		elog(ERROR, "Can't retrieve rangetable index for NULL string!");

	return result;

}

/*------------------------------------------------------------------------------
 * Retrieves the rule tuple.
 *------------------------------------------------------------------------------
 */
static char * retrieve_rule(Oid relOid,
							const char *ruleName,
							const bool noLock) 
{

	HeapTuple rule = NULL;
	char *result   = NULL;
	Assert((relOid != InvalidOid) && (ruleName != NULL));
	
	rule = SearchSysCache(RULERELNAME,
						  ObjectIdGetDatum(relOid),
						  PointerGetDatum(ruleName),
						  0, 0 );

	if (HeapTupleIsValid(rule)) 
	{

		Relation  rewrite_relation = NULL;
		TupleDesc rewrite_desc     = NULL;
		Datum     ruleAction;
		bool      isNull           = false;

		/*
		 * If NoLock is false we acquire a row exclusive here, since
		 * someone wants to work on the returned rule.
		 */
		
		if (noLock)
			rewrite_relation = heap_open(RewriteRelationId, NoLock);
		else
			rewrite_relation = heap_open(RewriteRelationId, RowExclusiveLock);

		rewrite_desc     = RelationGetDescr(rewrite_relation);
		ruleAction = heap_getattr(rule,
								  Anum_pg_rewrite_ev_action,
								  rewrite_desc,
								  &isNull);
		result = DatumGetCString(DirectFunctionCall1(textout,
													 ruleAction));
		heap_close(rewrite_relation, NoLock);

	}

	ReleaseSysCache(rule);

	return result;

}

/*------------------------------------------------------------------------------
 * Returns the RangeTblEntry starting at the specified offset. The
 * function can be used to iterate over the rtable list of the
 * specified select query tree. NOTE: the function only returns 
 * RangeTblEntry that does not match to a *NEW* or *OLD* RangeTblEntry.
 * Returns NULL if nothing found. The offset of the last returned
 * RTE is saved in offset.
 *
 * NOTE:
 * the returned offset always holds the last rte index + 1!
 *------------------------------------------------------------------------------
 */
static RangeTblEntry *
get_relation_RTE(const Query  *select,
				 unsigned int *offset) 
{

	RangeTblEntry *result = NULL;
	List          *list   = NULL;

	Assert((offset != NULL) && (select != NULL));

	if (select == NULL) 
		elog(ERROR, "Attempt to extract RTE from uninitialized query tree!");

	if ((*offset + 1) > list_length(select->rtable))
		return result;
	
	list = select->rtable;

	while (*offset <= list_length(list)) 
	{    
		RangeTblEntry *rte = rt_fetch(*offset, list);
		++(*offset);

		if (rte != NULL) 
		{

			if (rte->rtekind == RTE_RELATION) 
			{ 

				if ((strncmp(rte->eref->aliasname, "*NEW*", 5) == 0) ||
					(strncmp(rte->eref->aliasname, "*OLD*", 5) == 0) ) 
				{
					continue;
				} else
					return rte;

			}
		}    
	}

	return result;

}

/*------------------------------------------------------------------------------
 * Replaces the varnos if node is of type Var. The varno is 
 * replaced only if oldIndex matches. Set oldIndex
 * to 0 if you want _all_ varnos to be replaced with
 * rtIndex.
 * subQueryLevel gives the outer relation level replace_varnos_for_var
 * is allowed to change, and so only Var's with varlevelsup values that
 * match subQueryLevel are changed.
 * Returns true in case of success.
 *------------------------------------------------------------------------------
 */
static bool
replace_varnos_for_var( Node *node,
						Index rtIndex,
						Index oldIndex, 
						Index subQueryLevel ) 
{

	bool result  = false;

	Assert( PointerIsValid(node) );
	Assert( rtIndex > 0);

	if ( IsA( node, Var ) ) 
	{
			
		if ( ( ((Var *)node)->varno == oldIndex ) 
			 && ( oldIndex > 0 )
			 && ( ((Var *)node)->varlevelsup == subQueryLevel ) ) 
		{
					
			elog( DEBUG1, "Replacing varno %d to %d in sublevel %d, varattno %d, varlevelsup %d",
				  ((Var *)node)->varno,
				  rtIndex,
				  subQueryLevel, 
				  ((Var *)node)->varattno, 
				  ((Var *)node)->varlevelsup );
			
			((Var *)node)->varno = rtIndex;
			((Var *)node)->varnoold = rtIndex;
			result = true;
			
		} 
		else if ( ( oldIndex == 0 ) 
				  && ( ((Var *)node)->varlevelsup == subQueryLevel ) ) 
		{

			elog( DEBUG1, "Replacing varno %d to %d in sublevel %d, varattno %d, varlevelsup %d",
				  ((Var *)node)->varno,
				  rtIndex,
				  subQueryLevel, 
				  ((Var *)node)->varattno, 
				  ((Var *)node)->varlevelsup );
			
			((Var *)node)->varno = rtIndex;
			((Var *)node)->varnoold = rtIndex;
			result = true;
			
		}
		
	}
	
	return result;

}

static void replace_varattno_for_var( Node *node,
									  ViewDefColumnList tentries, 
									  Index subQueryLevel ) 
{

	Assert((PointerIsValid(node)) && (!ColumnListIsNull(tentries)));

	if IsA( node, Var ) 
	{

		TargetEntry *entry = ColumnListGet(tentries, 
										   ((Var *)node)->varattno - 1);
		
		if (entry == NULL)
			return;
		
		if ( (entry->resno != ((Var *)node)->varattno) 
			 && (((Var *)node)->varlevelsup == subQueryLevel) ) 
		{
			
			((Var *)node)->varattno = entry->resno;
			((Var *)node)->varoattno = entry->resno;
			
		}
		
	}
	
}

/*------------------------------------------------------------------------------
 * Walks down the view qualifications node and 
 * moves all varnos/varnoold to point to the *NEW*
 * RTE.
 *------------------------------------------------------------------------------
 */
static bool
view_def_qual_walker(Node *node,
					 struct ViewExprContext *context)
{

	if ( !PointerIsValid(node) ) 
		return false;

	replace_varnos_for_var( node, context->newRTE, 
							0, context->subQueryLevel );
	replace_varattno_for_var( node, context->tentries, 
							  context->subQueryLevel );
	
	switch (node->type)
	{

	case T_CaseExpr:
		
		/*
		 * fall through
		 */
		
	case T_NullTest:
		
		ChangeVarNodes( (Node *)node,
						context->baseRTE, 
						context->newRTE, 0 );
		
		break;

	case T_RelabelType:
		
		
		ChangeVarNodes( (Node *)((RelabelType *)node)->arg,	
						context->baseRTE, context->newRTE, 0 );
		
		
		break;
		
	case T_Query:		
	{

		bool result;

		/*
		 * Here comes the part where we dive into
		 * the subselect query tree. we need this to detect
		 * any outer relation references in this subquery tree, because
		 * they needs to be rewritten to be pinned at the *NEW*
		 * pseudorelation. Note that this is necessary only
		 * for outer references to *NEW* or *OLD*
		 *
		 * According to the comments in src/include/nodes/primnodes.h
		 * we can rely on the fact that the subselect node is a
		 * Query * ...
		 */
		
		++(context->subQueryLevel);
		result = query_tree_walker( (Query *)node,
									view_def_qual_walker, 
									(void *)context, 0 );
		--(context->subQueryLevel);
		
		return result;
		
	}
	
	default:

		break;
		
	}

	return expression_tree_walker( (Node *)node,
								   view_def_qual_walker,
								   (void *)context );

}

static void
form_vars_arrayref(ArrayRef    *array,
				   Index       newRTE,
				   ViewDefColumnList tentries) 
{

	Assert( !ColumnListIsNull(tentries)
			&& (newRTE > 0)
			&& (array != NULL) );

	if (array->refexpr != NULL) 
	{
	
		if (IsA(array->refexpr, Var))
		{
			
			Var *var = (Var *)(array->refexpr);
			TargetEntry *entry = ColumnListGet(tentries, 
											   var->varattno - 1);
		  
			/*
			 * Only replace if column order is reversed.
			 */
			if (entry != NULL) 
			{
				
				if (entry->resno != var->varattno) 
				{
					
					var->varattno  = entry->resno;
					var->varoattno = entry->resno;
					
				}
				
			}

			/*
			 * At least, varno should point to the
			 * *NEW* range table entry
			 */
			var->varno = newRTE;
			var->varnoold = newRTE;

		}

	}

	if (array->refassgnexpr != NULL) 
	{

		if (array->refassgnexpr->type == T_Var) 
		{

			Var *var = (Var *)(array->refassgnexpr);
			TargetEntry *entry = ColumnListGet(tentries, var->varattno - 1);

			/*
			 * Only replace if column order is reversed.
			 */
			if (entry != NULL) 
			{

				if (entry->resno != var->varattno) 
				{

					var->varattno  = entry->resno;
					var->varoattno = entry->resno;

				}

			}

			/*
			 * At least, varno should point to the
			 * *NEW* range table entry
			 */
			var->varno = newRTE;
			var->varnoold = newRTE;

		}

	}

}

/*------------------------------------------------------------------------------
 * Creates an equal operator expression for the specified Var-pointers.
 *------------------------------------------------------------------------------
 */
static OpExpr *
create_opexpr( Var *var_left,
			   Var *var_right ) 
{

	OpExpr *result = NULL;
	Form_pg_operator operator;
	Operator         tuple;
	Assert( ( var_left != NULL ) && ( var_right != NULL ) );

	tuple = equality_oper( var_left->vartype, true );
  
	if (!HeapTupleIsValid( tuple ) ) 
	{

		elog( ERROR, 
			  "could not find equal-operator for UPDATE/DELETE where condition!" );

	} 
	else 
	{

		operator = (Form_pg_operator) GETSTRUCT( tuple );
		result = makeNode( OpExpr );

		result->opno = HeapTupleGetOid( tuple );
		result->opfuncid = operator->oprcode;
		result->opresulttype = operator->oprresult;
		result->opretset = false;
    
		result->args = lappend( result->args, var_left );
		result->args = lappend( result->args, var_right );

	}

	ReleaseSysCache( tuple );

	return result;

}

/*------------------------------------------------------------------------------
 * Creates a expression tree for a WHERE clause and assigns the root node 
 * (first call to build_expression_tree() only) to the specified FromExpr of
 * the target query tree. If you simply want to build a boolean expression tree
 * without assigning it to a query tree, set from = NULL.
 *------------------------------------------------------------------------------
 */
static Node *
build_expression_tree( FromExpr *from,
					   Node     **anchor,
					   BoolExpr *expr,
					   OpExpr   *op ) 
{

	/*
	 * Already some nodes there?
	 */

	if ( *anchor != NULL ) 
	{

		expr->args = lappend( expr->args, op );
		((BoolExpr *)(*anchor))->args = lappend( ((BoolExpr *)(*anchor))->args, 
												 expr );
		*anchor = (Node *)expr;

	} 
	else 
	{

		/*
		 * Currently no nodes...
		 */

		BoolExpr *boolexpr = makeNode( BoolExpr );
		expr->args = lappend( expr->args, op );
		boolexpr->args = lappend( boolexpr->args, 
								  expr );

		*anchor = (Node *)boolexpr;

		if ( from != NULL )
			from->quals = *anchor;

	}

	return (*anchor);

}

/*------------------------------------------------------------------------------
 * Forms the WHERE clause for DELETE/UPDATE
 * rules targeted to the specified VIEW OID.
 *------------------------------------------------------------------------------
 */
static void
form_where_for_updrule( const Query    *select,
						FromExpr       **from,
						const Relation rel,
						Oid            baserel,
						Index          baserti,
						Index          oldrti ) 
{

	BoolExpr *expr           = NULL;
	Node     *anchor         = NULL;
	OpExpr   *op             = NULL;
	Form_pg_attribute *attrs = rel->rd_att->attrs;
	ListCell *cell           = NULL;

	Assert((*from != NULL) 
		   && (OidIsValid( baserel )) 
		   && (baserti > 0) 
		   && (oldrti > 0) 
		   && (rel != NULL));

	foreach(cell, select->targetList) 
	{

		TargetEntry *te   = (TargetEntry *)lfirst(cell);
		Var         *var1 = NULL;
		Var         *var2 = NULL;

		CaseExpr *newcase   = makeNode(CaseExpr); 
		CaseWhen *casewhen  = makeNode(CaseWhen);
		NullTest *nulltest1 = makeNode(NullTest);
		NullTest *nulltest2 = makeNode(NullTest);

		/*
		 * If te->expr holds no Var pointer, continue ...
		 */

		if (!IsA( te->expr, Var )) 
			continue;

		/*
		 * These are the new operands we had to check
		 * for equality.
		 * 
		 * For DELETE/UPDATE rules, var1 points to the *OLD*
		 * RTE, var2 references the base relation.
		 *
		 */

		var1 = copyObject((Var *)(te->expr));

		/*
		 * Look at varoattno to determine wether this
		 * attribute has a different location in the underlying
		 * base table. If true, retrieve the attribute from the
		 * base table and assign it to var2, else we have to do a
		 * simple copy to var1.
		 */

		if (var1->varoattno > 0) 
		{

			var2 = makeNode( Var );

			var2->varno    = baserti;
			var2->varnoold = baserti;
			var2->varattno = attrs[ var1->varoattno - 1 ]->attnum;    
			var2->vartype     = attrs[ var1->varoattno - 1 ]->atttypid;
			var2->vartypmod   = attrs[ var1->varoattno - 1 ]->atttypmod;
			var2->varlevelsup = var1->varlevelsup;
			var2->varnoold    = var2->varno;
			var2->varoattno   = var2->varattno;
      
		} 
		else 
		{

			var2 = copyObject( var1 );
			var2->varno    = baserti;
			var2->varnoold = baserti;

		}
    
		var1->varno       = oldrti;
		var1->varnoold    = oldrti;

		/*
		 * rewrite varattno of var2 to point to the right
		 * column in relation *OLD* or *NEW*
		 */
		var2->varattno = te->resorigcol;
		var2->varoattno = te->resorigcol;

		/*
		 * rewrite varattno of var1 to point to the
		 * right column in base relation
		 */
		var1->varattno = te->resno;
		var1->varoattno = te->resno;


		op             = create_opexpr( var1, var2 );
		expr           = makeNode( BoolExpr );
		expr->boolop   = AND_EXPR;

		/*
		 * Finally, create the OpExpr node, as part of
		 * a CaseExpr and include the OpExpr as part of the Case
		 * for managing NULL's we will do this everytime.
		 * That way we will have no problem with:
		 *
		 * ALTER TABLE ... ALTER COLUMN ... DROP NOT NULL;
		 */

		nulltest1->arg = (Expr *)var1;
		nulltest1->nulltesttype = IS_NOT_NULL;

		nulltest2->arg = (Expr *)var2;
		nulltest2->nulltesttype = IS_NULL;

		casewhen->expr = (Expr *)nulltest1;
		casewhen->result = (Expr *)op;

		newcase->args = list_make1(casewhen);
		newcase->defresult = (Expr *) nulltest2;

		op = copyObject(newcase);

		anchor = build_expression_tree( *from, 
										(Node **)&anchor, 
										expr, 
										op );

	}

}

/*------------------------------------------------------------------------------
 * Replaces the varnos for the specified targetlist
 * to rtIndex
 *------------------------------------------------------------------------------
 */
static int
replace_varnos_for_target( List        *targetList,
						   ViewDefColumnList tentries,
						   Index       rtIndex ) 
{

	int result        = 0;
	ListCell    *cell = NULL;
	TargetEntry *te   = NULL;

	Assert( targetList != NIL );
	Assert( rtIndex > 0 );

	foreach( cell, targetList ) 
	{

		Node *node = (Node *)lfirst(cell);

		if ( IsA( node, TargetEntry) ) 
		{

			te = (TargetEntry *)node;

			if ( IsA( te->expr, Var ) ) 
			{

				((Var *)(te->expr))->varno = rtIndex;
				++result;

			}

		} 
		else if (IsA( te->expr, ArrayRef)) 
		{

			/*
			 * Things are getting complicater here. We have
			 * found an array subscripting operation. It's
			 * necessary to examine all varno's found in this
			 * operation to make sure, we're getting right.
			 * This covers cases where a view selects a 
			 * single index or complete array from a base 
			 * table or view.
			 */

			ArrayRef *array = (ArrayRef *)(te->expr);
			
			if ( array->refupperindexpr != NIL ) 
			{

				/*
				 * Look at expressions that evaluate upper
				 * array indexes. Make sure all varno's are modified
				 * This is done while diving into the expression
				 * tree recursively.
				 */

				replace_varnos_for_target( array->refupperindexpr,
										   tentries,
										   rtIndex );

			}

			if ( array->reflowerindexpr != NIL ) 
			{

				/*
				 * The same here as for refupperindexpr ...
				 */

				replace_varnos_for_target( array->reflowerindexpr,
										   tentries,
										   rtIndex );

			}

			/*
			 * At least look at the expressions ...
			 */

			if (array->refexpr != NULL) 
			{

				if (IsA(array->refexpr, Var))
					form_vars_arrayref( array,
										rtIndex,
										tentries );

			}

			if (array->refassgnexpr != NULL) 
			{

				if (IsA(array->refassgnexpr, Var))
					form_vars_arrayref( array,
										rtIndex,
										tentries );

			}

		}

	}

	return result;

}

/*------------------------------------------------------------------------------
 * Remove all columns from the specified targetList that aren't 'real' table
 * columns. So all system columns, FuncExpr et al. are removed.
 * The function returns false, if no more columns are left (that is, the 
 * view target list doesn't hold any 'real' columns...)
 *------------------------------------------------------------------------------
 */
static bool
remove_sys_columns( List *targetList,
		            List *sourceList ) 
{

	ListCell *cell   = NULL;
	int      r_count = 0;
	int      l_count = list_length(targetList);

	foreach( cell, sourceList ) 
	{

		Node  *node = (Node *)lfirst(cell);
		Index rt_offset = 0;

		if ( IsA( node, TargetEntry ) ) 
		{

			TargetEntry *te = (TargetEntry *)node;

			if ( IsA( te->expr, Var ) ) 
			{

				/*
				 * varattno's with negative values indicates syscolls.
				 * Ignoring them means removing from the target list...
				 */

				Var *var = (Var *)te->expr;

				if (var->varattno < 0) 
				{

					targetList = list_delete( targetList, node );
					elog( DEBUG1, "system columns aren't updatable" );

					++rt_offset;
					++r_count;

				} 
			} 
			else 
			{

/* 				if (IsA( te->expr, ArrayRef ) )  */
/* 				{ */

/* 					targetList = list_delete( targetList, node ); */
/* 					elog( NOTICE, "array expressions aren't updatable yet" ); */
/* 					++rt_offset; */
/* 					++r_count; */


/* 				} */

				if (IsA( te->expr, FuncExpr ) ) 
				{

					targetList = list_delete( targetList, node );
					elog( DEBUG1, "function expressions aren't updatable" );
					++rt_offset;
					++r_count;

				}

			}

			te->resno -= rt_offset;


		}
		
	}

	if (r_count >= l_count)
		return false;
	else
		return true;

}

/*------------------------------------------------------------------------------
 * Adds RTEs to form a query tree. 
 * select has to be a valid initialized
 * view definition query tree (the function assumes that
 * this query has passed the checkTree() function).
 *------------------------------------------------------------------------------
 */
static void form_query( const Query *select,
						Query       *query,
						ViewDefColumnList tentries,
						bool        copyTargetList) 
{

	RangeTblEntry *rte    = NULL;
	ListCell      *cell   = NULL;
	Oid            reloid = InvalidOid;

	Assert( ( select != NULL ) 
			&& ( query != NULL )
			&& ( tentries != NULL ) );

	/*
	 * Copy the range table entries
	 */
	query->rtable = copyObject( select->rtable );

	/*
	 * Prepare the other stuff
	 */

	query->canSetTag = true;
	query->jointree = makeNode( FromExpr );

	/*
	 * Those entries in the range table which has the inFromCl flag
	 * set must be changed.
	 */

	foreach( cell, query->rtable ) 
	{
    
		Node *node = (Node *)lfirst( cell );
		if (IsA( node, RangeTblEntry )) 
		{
      
			((RangeTblEntry *)node)->inFromCl = false;
			((RangeTblEntry *)node)->inh = false;
      
		}
    
	}

	/*
	 * Set result relation to the base relation.
	 * Since we currently only support SQL92 views, we
	 * simply extract the one relation, which isn't labeled
	 * as *OLD* or *NEW*
	 */

	reloid = get_reloid_from_select( select, &rte );
  
	if ( reloid > 0 ) 
	{

		query->resultRelation = get_rtindex_for_rel( query->rtable,
													 rte->eref->aliasname );
    
	} 
	else
		elog( ERROR, "Cannot retrieve OID for base relation!" );

	Assert( query->resultRelation > 0 );

	if (copyTargetList) 
	{

		Index new_rtindex = 0;
		List *tempList = copyObject( select->targetList );
		if (!remove_sys_columns( select->targetList,
								 tempList ))
		{
			elog(ERROR, "no updatable columns in view!");
		}

		/*
		 * Copy all target entries.
		 */
		query->targetList = copyObject( select->targetList );

		/*
		 * We have to replace all varnos to point to the
		 * *NEW* node in all targetentry expressions.
		 */
		new_rtindex = get_rtindex_for_rel( query->rtable,
										   "*NEW*" );
		replace_varnos_for_target( query->targetList, tentries, 
								   new_rtindex );
    
	}

}

static void 
form_targetentry_for_update( int2     attnum,
							 Form_pg_attribute attrs,
							 Oid         baserel,
							 Expr        *expr,
							 TargetEntry *te_update ) {

	/*
	 * Try first if this is an array subscripting
	 * operation. If true, dive recursively into
	 * the subscripting tree examining all varnos.
	 */

	if (expr->type == T_ArrayRef) 
	{

		ArrayRef *array = (ArrayRef *)expr;

		if (array->refassgnexpr != NULL) 
		{

			form_targetentry_for_update( attnum,
										 attrs,
										 baserel,
										 array->refassgnexpr,
										 te_update );
 
		}

		if (array->refupperindexpr != NIL) 
		{

			ListCell *cell = NULL;
			foreach(cell, array->refupperindexpr) 
			{
				
				Expr *subexpr = (Expr *)lfirst(cell);
				form_targetentry_for_update( attnum,
											 attrs,
											 baserel,
											 subexpr,
											 te_update );

			}

		}

		if (array->reflowerindexpr != NIL) 
		{

			ListCell *cell = NULL;
			foreach(cell, array->reflowerindexpr) 
			{

				Expr *subexpr = (Expr *)lfirst(cell);
				form_targetentry_for_update( attnum,
											 attrs,
											 baserel,
											 subexpr,
											 te_update );

			}

		}

		if (array->refexpr != NULL) 
		{

			form_targetentry_for_update( attnum,
										 attrs,
										 baserel,
										 array->refexpr,
										 te_update );

		}

	} 
	else if (expr->type == T_Var)
	{
	
		/*
		 * NOTE:
		 *   upd_var points to the column in the view
		 */
		
		Var *upd_var       = (Var *)(te_update->expr);
		upd_var->varattno  = te_update->resno;
		
		upd_var->varoattno = te_update->resno;
		upd_var->vartype   = attrs->atttypid;
		upd_var->vartypmod = attrs->atttypmod;
		upd_var->varnoold  = upd_var->varno;

		te_update->resno           = attnum;
		te_update->resname         
			= pstrdup( get_attname( baserel, 
									attnum ) );		
		te_update->ressortgroupref = 0;
		te_update->resorigcol      = 0;
		te_update->resorigtbl      = 0;
		te_update->resjunk         = false;
		
	}

}

/*------------------------------------------------------------------------------
 * Build the target list for an view UPDATE
 * rule.
 *------------------------------------------------------------------------------
 */
static void
build_update_target_list(  const Query *update,
						   const Query *select,
						   ViewDefColumnList tentries,
						   Oid         baserel,
						   Relation    rel ) 
{

	ListCell          *cell = NULL;
	Form_pg_attribute attrs = NULL;
	Index             listIndex = 0;

	Assert( (update != NULL)
			&& (rel != NULL)
			&& (select != NULL)
			&& (tentries != NULL )	  
			&& (HeapTupleIsValid(rel)) );

	/*
	 * We loop over the targetlist of the select query
	 * tree. 
	 */

	foreach(cell, select->targetList) 
	{

		TargetEntry *entry     = (TargetEntry *)lfirst(cell);
		int attindex           = 0;
		TargetEntry *upd_entry = NULL;
		
		if (entry->resorigcol > 0) 
		{			
			/*
			 * This column seems to have a different order than
			 * in the base table.
			 * We get the attribute from the base relation referenced
			 * by rel and create a new resdom. This new result domain
			 * is then assigned instead of the old one.
			 */	     
			attindex = entry->resorigcol;			
		} 
		else 
		{			
			/* 	attindex = var->varattno; */
			attindex = entry->resno;			
		}
		
		upd_entry = list_nth( update->targetList, listIndex );
		
		if (upd_entry == NULL)
			continue;
		
		attrs = rel->rd_att->attrs[ attindex - 1 ];

		form_targetentry_for_update( attindex,
									 attrs,
									 baserel,
									 upd_entry->expr,
									 upd_entry );
									 
		listIndex++;
		
	}

}

/*------------------------------------------------------------------------------
 * Returns the pg_proc tuple for the view definition
 * qualification check function. The function raises
 * always an error, if no tuple can be found.
 *------------------------------------------------------------------------------
 */
static struct ViewQualCheckFunc
get_view_qualification_func( const char *name ) 
{

	struct ViewQualCheckFunc result;
	HeapTuple    heapProc  = NULL;

	Value *funcname             = makeString( (char *)name );
	Value *typname              = makeString( "bool" );
	TypeName     *type          = makeNode( TypeName );
	List         *funcname_list = list_make1( funcname );
	List         *typnames_list = list_make1( type );

	Assert( name != NULL );

	/*
	 * Create typename list
	 */
	type->names = list_make1( typname );

	/*
	 * Lookup function name for view qualification check function
	 */

	result.procoid = LookupFuncNameTypeNames( funcname_list,
											  typnames_list,
											  FALSE );

	if ( OidIsValid( result.procoid ) ) 
	{

		elog( DEBUG1, "Got OID %u for check function %s!", 
			  result.procoid, PG_VIEW_UPDATE_CONTROL );

		heapProc = SearchSysCacheCopy( PROCOID,
									   ObjectIdGetDatum( result.procoid ),
									   0, 0, 0 );

		if (!HeapTupleIsValid( heapProc )) 
		{

			elog( ERROR, "Could not retrieve tuple for OID %u!",
				  result.procoid );

		} 
		else 
		{

			result.heap  = heapProc;
			result.tuple = (Form_pg_proc) GETSTRUCT( heapProc);

		}

	} else
		elog( ERROR, "Can't find view qualification check function %s!",
			  PG_VIEW_UPDATE_CONTROL );

	return result;

}

/*------------------------------------------------------------------------------
 * Create a rule qualification from the specified
 * view definition query tree. This qualification are copied
 * to the stmt's ev_action query tree and evaluated to 
 * check a view definition's qualification. We get
 * a WITH CHECK OPTION behavior by creating a
 * conditional rule here. The view definition qualification
 * is evaluated by the PG_VIEW_UPDATE_CONTROL function.
 * The function returns the function node or null.
 *------------------------------------------------------------------------------
 */
static FuncExpr *
create_qual_for_rule( RuleStmt    *stmt,
					  const Query *select,
					  ViewDefColumnList tentries,
					  Index       newRTE,
					  Index       oldRTE ) 
{

	FuncExpr *   func    = makeNode( FuncExpr );
	Index        baseRTE = 0;
	struct ViewQualCheckFunc proc;
	struct ViewExprContext context;

	Assert( stmt != NULL );
	Assert( list_length( stmt->actions ) >= 1 );
	Assert( select != NULL );
	Assert( (newRTE > 0) && (oldRTE > 0) );

	if ( select->jointree->quals == NULL ) 
	{
		/* nothing to do here */
		return NULL;
	}

	/*
	 * Retrieve the func id by name.
	 */
  
	proc = get_view_qualification_func( PG_VIEW_UPDATE_CONTROL );

	/*
	 * Only reached if no error has occured (should not happen)
	 */
	func->funcid         = proc.procoid;
	func->funcresulttype = proc.tuple->prorettype;
	func->funcretset     = false;
	func->funcformat     = 0;

	/*
	 * Move the view qualification over to the RuleStmt
	 */
	func->args = list_make1( copyObject( select->jointree->quals ) );
	stmt->whereClause = (Node *)func;

	/*
	 * We have to dig into the WHERE clause and
	 * grep out all varnos that have to point to the *NEW* or
	 * *OLD*  RTE.
	 */
	baseRTE = 2;
	get_relation_RTE( select, &baseRTE );
	Assert( baseRTE > 2 );

	context.newRTE        = newRTE;
	context.baseRTE       = baseRTE - 1;
	context.oldRTE        = oldRTE;
	context.subQueryLevel = 0;
	context.tentries      = tentries;

	expression_tree_walker( (Node *)func, 
							view_def_qual_walker, 
							(void *)&context );

	/*
	 * free stuff
	 */
	heap_freetuple( proc.heap );

	return func;

}

/*------------------------------------------------------------------------------
 * Creates a NOTHING rule for INSERTs.
 *------------------------------------------------------------------------------
 */
static RuleStmt *
create_nothing_rule( Query          *action,
					 const RangeVar *var,
					 const char     *rulename,
					 CmdType        cmdType,
					 bool           replace ) {

	RuleStmt *rule = NULL;

	Assert( action != NULL );

	if ( action == NULL )
		elog( ERROR, "can't create NOTHING rule from uninitialized query tree!" );
	else 
	{

		rule = makeNode( RuleStmt );

		rule->relation = copyObject( (RangeVar *)var );

		rule->rulename    = pstrdup( rulename  );
		rule->whereClause = NULL;
		rule->event       = cmdType;
		rule->instead     = TRUE;
		rule->actions     = list_make1( action );
		rule->replace     = replace;

	}

	return rule;

}

/*------------------------------------------------------------------------------
 * Creates a INSERT RuleStmt structure with the specified arguments
 *------------------------------------------------------------------------------
 */
static RuleStmt *
create_insert_rule( Query          *insert,
					const RangeVar *view,
					bool           replace ) 
{

	RuleStmt *rule     = makeNode( RuleStmt );
  
	/*
	 * Set target relation for rule.
	 */
	rule->relation = copyObject( (RangeVar *)view );

	rule->rulename = pstrdup( INSERTRULENAME );
	rule->whereClause = NULL;
	rule->event = CMD_INSERT;
	rule->instead = true;
	rule->actions = list_make1( insert );
	rule->replace = replace;

	return rule;

}

/*------------------------------------------------------------------------------
 * Creates a UPDATE RuleStmt structure with the specified arguments
 *------------------------------------------------------------------------------
 */
static RuleStmt *
create_update_rule( Query          *update,
					const RangeVar *view,
					bool           replace ) 
{

	RuleStmt *rule     = makeNode( RuleStmt );

	/*
	 * Set target relation for rule.
	 */
	rule->relation = copyObject( (RangeVar *)view );

	rule->rulename = pstrdup( UPDATERULENAME );
	rule->whereClause = NULL;
	rule->event = CMD_UPDATE;
	rule->instead = true;
	rule->actions = list_make1( update );
	rule->replace = replace;

	return rule;

}

/*------------------------------------------------------------------------------
 * Creates a DELETE RuleStmt structure with the specified arguments
 *------------------------------------------------------------------------------
 */
static RuleStmt *
create_delete_rule( Query          *delete,
					const RangeVar *view,
					bool           replace ) 
{

	RuleStmt *rule     = makeNode( RuleStmt );
  
	/*
	 * Set target relation for rule.
	 */
	rule->relation = copyObject( (RangeVar *)view );

	rule->rulename = pstrdup( DELETERULENAME );
	rule->whereClause = NULL;
	rule->event = CMD_DELETE;
	rule->instead = true;
	rule->actions = list_make1( delete );
	rule->replace = replace;

	return rule;

}

/*------------------------------------------------------------------------------
 * Takes the specified ev_action buffer and transforms it into
 * a query tree. Returns NULL if an error has occured or the Query Tree
 * stored in buf
 *------------------------------------------------------------------------------
 */
static Query *
transform_to_QueryTree( StringInfoData buf ) 
{

	List  *ev_action = NULL;
	Query *result = NULL;

	if ( buf.len < 1 ) 
		elog( ERROR, "Cannot transform ev_action to query tree!" );
	else 
	{

		ev_action = stringToNode( buf.data );
    
		if ( list_length(ev_action) == 1 ) 
		{

			Query    *query;
			query = (Query *)linitial( ev_action );
			result = query;

		} 
		else if (list_length(ev_action) > 1 ) 
		{

			elog( ERROR, "Could not transform ev_action to query tree!" ); 

		}

	}

	return result;

}

static Oid
hasRule( Oid        view,
		 const char *name ) 
{

	Oid result = InvalidOid;
	Assert( OidIsValid( view ) );

	if (OidIsValid( view )) 
	{

		HeapTuple tuple;

		tuple = SearchSysCache( RULERELNAME,
								ObjectIdGetDatum( view ),
								PointerGetDatum( name ),
								0, 0 );

		if ( HeapTupleIsValid( tuple ) ) 
		{

			result = HeapTupleGetOid( tuple );

			ReleaseSysCache( tuple );

		}

	}

	return result;

}

/*
 * Check if we can create the insert rule for this view.
 * If we can create DEFAULT values for all columns that needed.
 *
 * Up to now the only reason i can see to not create the rule is if the
 * view does not contains all not null without default fields of the relation.
 */
static bool
isInsertable(const Query *select, const Relation rel) 
{

	const int2 natts = RelationGetNumberOfAttributes(rel);
	int i;
	int num_sys_colls = 0;
	bool attr_ok[natts];
	ListCell *cell = NULL;
	bool ret_value = true;

	/*
	 * Initializing attr_ok array with false values
	 */
	for (i=0; i < natts; i++) attr_ok[i] = false;

	/*
	 * We loop over the targetlist of the select query
	 * tree. 
	 */
	foreach(cell, select->targetList) 
	{
		TargetEntry *entry     = (TargetEntry *)lfirst(cell);

		if (! IsA( entry->expr, Var ) &&
			! IsA( entry->expr, ArrayRef ) )
		{
			
			/*
			 * Count non-Var expressions. We need to check if there
			 * are 'normal' table columns left if we gonna to remove
			 * all others with remove_sys_columns()
			 */
			
			num_sys_colls++;
			
		} 
		else if (entry->resorigcol <= 0)
		{
				 
			num_sys_colls++;
			
			/*
			 * If resorigcol < 0 then this is a system 
			 * col and will be ignored
			 * Or
			 * If resorigcol = 0 then it is not base rel 
			 * col and will be ignored 
			 * The system cols will be removed from the targetlist later
			 */
			continue;

		}

		/*
		 * This is just to mark the column so we know it's in the 
		 * view definition
		 * At this point it doesn't matter if the column is not null
		 * nor if it has default values
		 */
		attr_ok[entry->resorigcol-1] = true;

	}

	for (i=0; i < natts; i++) 
	{
		/*
		 * If the column is in the view definition then no problem
		 */
		if (attr_ok[i]) continue;

		/*
		 * Ignore dropped columns 
		 */
		if (rel->rd_att->attrs[i]->attisdropped) continue; 
		
		/*
		 * If there is one column that is not in the view definition
		 * and not allow nulls and not has a default value then the 
		 * view is not insertable 
		 */
		if ((rel->rd_att->attrs[i]->attnotnull) && 
	     	   !(rel->rd_att->attrs[i]->atthasdef)) 
		{
			return false;			
		}

	}

	/*
	 * At least, check if a call to remove_sys_columns() will leave
	 * an empty target list.
	 */
	
	if (list_length(select->targetList) - num_sys_colls <= 0)
		return false;
	
	return ret_value;	
}

/*------------------------------------------------------------------------------
 * Published functions
 * -----------------------------------------------------------------------------
 */

char
makeViewCheckOption(bool check,
	
				bool cascaded) {
	
	char result = NO_OPTION_IMPLICIT;

	if (check)
	{
		if (cascaded)
			result = CASCADED_OPTION_IMPLICIT;
		else
			result = LOCAL_OPTION_IMPLICIT;
	}

	return result;

}

/*------------------------------------------------------------------------------
 * Returns the base table(s) for the specified relation OID. The result is a list
 * of all possible base table(s) the given view is based on.
 * -----------------------------------------------------------------------------
 */
extern void
get_base_relations( struct ViewBaseRelation *tree,
	                List   **baserelations ) 
{

	if ( tree != NULL ) 
	{

		ListCell *acell = NULL;

		/*
		 * nothing to do?
		 */
		if ( tree->defs == NIL ) 
			return;

		foreach( acell, tree->defs ) 
		{

			ListCell *bcell = NULL;
			struct ViewBaseRelation *relations 
				= ( struct ViewBaseRelation *)lfirst(acell);


			/*
			 * current children holds a base table?
			 */
			
			foreach( bcell, relations->defs ) 
			{

				struct ViewBaseRelationItem *item
					= (struct ViewBaseRelationItem *)lfirst(bcell);

				if ( item->isView == FALSE ) 
				{
					elog( DEBUG1, "Found base relation %d", item->reloid );
					*baserelations = lappend( *baserelations, item );
				}

			}

		}

	}

}

/*------------------------------------------------------------------------------
 * Examines the columns by the current view and initializes the
 * lookup table for all rearranged columns in base relations. The function
 * requires a relation tree initialized by get_base_base_relations().
 *------------------------------------------------------------------------------
 */

extern void
read_rearranged_colls( struct ViewBaseRelation *tree ) 
{

	Assert( (tree != NULL) );

	if (tree->defs != NIL) 
	{

		/*
		 * Traverse the relation tree and look on all
		 * base relations for reversed column order in their
		 * target lists. We have to perform a look-ahead-read on
		 * the tree, because we need to know how much columns
		 * the next base relation has, to allocate enough memory
		 * in tentries.
		 *
		 * Note that if only one base relation (a "real" table, not a
		 * view) exists, we have nothing to do, because this base relation
		 * cannot have a reversed column order caused by a view
		 * definition query.
		 */		
			
		int i           = 0;
		int num_items   = list_length( tree->defs );
		
		for ( i = ( num_items - 1 ); i > 0; --i ) 
		{

			int nattr       = 0;
			struct ViewBaseRelationItem *item_current = NULL;
			struct ViewBaseRelationItem *item_next    = NULL;
			struct ViewBaseRelation *current
				= (struct ViewBaseRelation *)list_nth( tree->defs, i );

			/*
			 * We look ahead for the next base relation. We can do
			 * this here safely, because the loop condition
			 * terminates before reaching the list head.
			 */
			
			struct ViewBaseRelation *next
				= (struct ViewBaseRelation *)list_nth( tree->defs, (i - 1) );

			/*
			 * Note that the code currently requires a SQL92 only
			 * relation tree. This means we handle one base relation
			 * per loop, only.
			 */

			Assert( next != NULL );
			Assert( current != NULL );
			Assert( list_length( next->defs ) == 1 );
			Assert( list_length( current->defs ) == 1 );

			item_current 
				= (struct ViewBaseRelationItem *)list_nth( current->defs, 0 );
			item_next
				= (struct ViewBaseRelationItem *)list_nth( next->defs, 0 );

			/*
			 * Allocate tentries buffer
			 */
			
			nattr = get_relnatts( item_next->reloid );

			ColumnListCreate(item_current->tentries, nattr);

			copyReversedTargetEntryPtr( item_current->rule->targetList,
										item_current->tentries );

		}

	}

	return;	

}

/*------------------------------------------------------------------------------
 * Retrieves all relations from the
 * view that can be considered a "base relation".
 * The function returns a list that holds lists of all
 * relation OIDs found for the view. The list is filled top down,
 * that means the head of the list holds the relations
 * for the "highest" view in the tree.
 *
 * We get then a tree layout as follows,
 * consider this view definition tree where
 * each node is a relation the above node is based
 * on:
 *
 *                         1
 *                        / \
 *                       2   3
 *                      / \   \
 *                     4   5   6
 *                        /
 *                       7
 *
 * The function will then return a list with the following layout:
 *
 * Listindex          Node(s)
 * --------------------------
 * 1                  7
 * 2                  4 5 6
 * 3                  2 3
 *
 * As you can see in the table, the topmost relations 
 * are saved as the *last* entries in the list.
 *
 * NOTE: The top node (the view relation itself) is not saved.
 *------------------------------------------------------------------------------
 */
extern void
get_base_base_relations( const Query *view,
						 Oid         baserel,
						 List        **list ) 
{

	if ( view != NULL ) 
	{
    
		RangeTblEntry *entry              = NULL;
		unsigned int  offset              = 1;
		struct ViewBaseRelationItem *item = NULL;
		
		struct ViewBaseRelation *childRel = 
			(struct ViewBaseRelation *)palloc(sizeof(struct ViewBaseRelation));
		
		childRel->defs = NIL;
		childRel->fatherRelation = baserel;
		
		/*
		 * Get all OIDs from the RTE list of view.
		 */
		
		while ((entry = get_relation_RTE( view,
										  &offset )) != NULL) 
		{
			
			/*
			 * Is this really a view or relation?
			 */
			HeapTuple tuple = SearchSysCacheCopy( RELOID,
												  ObjectIdGetDatum( entry->relid ),
												  0, 0, 0 );
			
			if (HeapTupleIsValid(tuple)) 
			{
				
				Form_pg_class relation = (Form_pg_class)GETSTRUCT(tuple);
				Oid itemoid = HeapTupleGetOid(tuple);
				char *buf = NULL;

				if ( (relation->relkind == RELKIND_RELATION) 
					 || (relation->relkind == RELKIND_VIEW) ) 
				{
					
					item = (struct ViewBaseRelationItem *)
						palloc(sizeof(struct ViewBaseRelationItem));
					
					item->reloid = itemoid;
					item->rel    = relation;
					item->rule   = NULL;
					item->childs = NIL;
					
					if (relation->relkind == RELKIND_VIEW) 
					{

						item->isView = true;
						
						/*
						 * Get the rule _RETURN expression tree for the specified
						 * relation OID. We need this, to dig deeper into the
						 * view base relation tree. Please note that
						 * we acquire a row exclusive lock (passing
						 * false to retrieve_fule()) on the retrieving rule, 
						 * since we rely heavily on its	current structure.
						 */
						
						buf  = retrieve_rule(item->reloid,
											 ViewSelectRuleName, false );

					} 
					else 
					{
						item->rule = NULL;
						item->isView = FALSE;
					}

					childRel->defs = lappend( childRel->defs, item );

					elog( DEBUG1, "extracted OID %u for relation tree",
						  item->reloid );
					
					if ( (item->rel->relkind == RELKIND_VIEW) 
						 && (buf != NULL) )
					{
						
						/*
						 * Dive into child relations recursively.
						 */						
						StringInfoData bufStrInfo;

						initStringInfo( &bufStrInfo );
						appendStringInfo( &bufStrInfo, buf );
						item->rule = transform_to_QueryTree( bufStrInfo );
						
						/* buffer no longer needed */
						pfree( bufStrInfo.data );

						get_base_base_relations( item->rule,
												 item->reloid,
												 list );
						
					}
					
				}
				
			}
			
		}
		
		if (childRel->defs != NIL)
			*list = lappend( *list, childRel );

	}
	
	return;
	
}

extern void
copyReversedTargetEntryPtr( List              *targetList,
							ViewDefColumnList targets ) 
{

	ListCell *cell = NULL;
	Assert( !ColumnListIsNull(targets) && (targetList != NIL) );

	/*
	 * NOTE: we only reassign pointers.
	 */

	foreach(cell, targetList) 
	{

		Node *node = (Node *)lfirst(cell);

		if (IsA(node, TargetEntry)) 
		{

			/*
			 * Look at the resdom's resorigcol to determine
			 * wether this is a reversed column (means, it has a 
			 * different column number than the underlying base table).
			 */
			TargetEntry *entry = (TargetEntry *)node;

			if (!IsA(entry->expr, Var))
				/* nothing todo here */
				continue;

			if ( (entry->resorigcol > 0) 
				 && ( (entry->resno != entry->resorigcol) ) ) 
			{

				/*
				 * Save this TargetEntry to the appropiate place 
				 * in the lookup table.
				 * Save only if not already occupied (this could 
				 * happen if the column is specified more than
				 * one time in the view definition).
				 */
				if (ColumnListItemIsNull(targets, entry->resorigcol - 1)) 
				{
					ColumnListSet(targets, entry->resorigcol - 1, entry);
				}

			}

		}

	}

}

/*------------------------------------------------------------------------------
 * Transforms the specified view definition into
 * a DELETE rule. 
 * Note: The function assumes that the specified query tree
 * has passed the checkTree() function.
 *------------------------------------------------------------------------------
 */
extern Query *
transform_select_to_delete( const Query       *select,
							const Relation    rel,
							const RangeVar    *var,
						    ViewDefColumnList tentries,
							bool              checkOption,
							bool              checkCascade ) 
{
	
	Query         *delete   = NULL;
	RuleStmt      *nothing  = NULL;
	RuleStmt      *rule     = NULL;
	Oid           baserel   = 0;
	RangeTblEntry *entry    = NULL;
	RangeTblEntry *viewrte  = NULL;
	char          *viewname = NULL;
	bool          replace_nothing   = FALSE;
	bool          replace_delete    = FALSE;

	Assert( (tentries != NULL) 
			&& (rel != NULL)
			&& (var != NULL)
			&& (select != NULL) );
	
	baserel = get_reloid_from_select( select, &entry );
		
	if ( baserel > 0 ) 
	{
		
		Index oldrti         = 0;
		Index baserti        = 0;			
		RangeTblRef *oldref  = NULL;
		RangeTblRef *baseref = NULL;
		Query *gen_delete    = NULL;
		Index rtindex        = 0;
		char  option         = makeViewCheckOption(checkOption,
												   checkCascade);
		
		delete              = makeNode( Query );
		delete->commandType = CMD_DELETE;
		
		/* 
		 * Catch the elog() from form_query(). We don't want
		 * to error out here, otherwise the whole CREATE VIEW
		 * command will fail.
		 */
		PG_TRY();
		{
			form_query( select, delete, tentries, false );
		}
		PG_CATCH();
		{
			elog(DEBUG1, "could not create DELETE update rule");
			return NULL;
		}
		PG_END_TRY();
		
		/*
		 * Pass this generic tree to create_nothing_rule(), 
		 * it's sufficient to create a INSTEAD DO NOTHING
		 * rule for deletion.
		 * 
		 * This is required, because we can get conditional rules
		 * if the checkOption is specified. We only save this generic tree
		 * here, the work is done at the end of this function.
		 */
		
		gen_delete = copyObject( delete );
		gen_delete->commandType = CMD_NOTHING;
		
		/*
		 * form_query() has prepared the jointree of the
		 * new DELETE rule. However, a DELETE rule needs range table 
		 * references to *OLD* and base relation RTEs.
		 */
		
		baserti = get_rtindex_for_rel( delete->rtable,
									   entry->eref->aliasname );
		
		oldrti = get_rtindex_for_rel( delete->rtable,
									  "*OLD*" );
		
		Assert( ( oldrti > 0 ) && ( baserti > 0 ) );
		
		oldref = makeNode( RangeTblRef );
		oldref->rtindex = oldrti;
		
		baseref = makeNode( RangeTblRef );
		baseref->rtindex = baserti;
		
		delete->jointree->fromlist = list_make1( baseref );
		delete->jointree->fromlist = lappend( delete->jointree->fromlist,
											  oldref );
		
		/*
		 * Create the WHERE condition qualification for the rule action.
		 */
		form_where_for_updrule( select,
								&(delete->jointree),
								rel,
								baserel,
								baserti,
								oldrti );
		
		/*
		 * Get the relation name for the view
		 * 
		 * ???? Hmm, sure we can rely on *NEW* here ????
		 */
		rtindex = get_rtindex_for_rel( delete->rtable, "*NEW*" );      
		Assert( rtindex > 0 );
		viewrte = rt_fetch( rtindex, delete->rtable );
		viewname = get_rel_name( viewrte->relid );
		
		elog( DEBUG1, "Viewname for relid %u is %s", viewrte->relid, viewname );
		
		/*
		 * Set ACL bit
		 */
		entry->requiredPerms |= ACL_DELETE;
		
		/*
		 * DELETE rule has to be replaced?
		 */
		if ( OidIsValid(hasRule( viewrte->relid, DELETERULENAME ) ) ) 
		{				
			replace_delete = TRUE;	
		} 
		else 
		{	
			replace_delete = FALSE;
		}
		
		/*
		 * NOTHING rule has to be replaced?
		 */
		if ( OidIsValid(hasRule( viewrte->relid, NOTHING_DELETERULENAME ) ) ) 
		{
			replace_nothing = TRUE;
		} 
		else 
		{
			replace_nothing = FALSE;
		}
		
		nothing = create_nothing_rule( gen_delete,
									   var,
									   NOTHING_DELETERULENAME,
									   CMD_DELETE,
									   replace_nothing );
		
		rule = create_delete_rule( delete,
								   var,
								   replace_delete );
		
		/*
		 * Enter rule qualification creation. The DELETE rule
		 * has to reference the *OLD* range table entry.
		 */
		rtindex = get_rtindex_for_rel( delete->rtable, "*OLD*" );
		
		if (checkOption)
			create_qual_for_rule( rule, select, tentries, rtindex, baserti );
		
		DefineQueryRewrite( nothing, NO_OPTION_IMPLICIT );
		DefineQueryRewrite( rule, option );
		
	}
	
	return delete;
	
}

/*------------------------------------------------------------------------------
 * Transforms the specified SELECT query tree into
 * an equivalent UPDATE statement. Note:
 * The function assumes that the specified query tree
 * has passed the checkTree() function before.
 *------------------------------------------------------------------------------
 */
extern Query *
transform_select_to_update( const Query       *select,
							const Relation    rel,
							const RangeVar    *var,
							ViewDefColumnList tentries,
							bool              checkOption,
							bool              checkCascade ) 
{

	Query         *update   = NULL;
	Oid           baserel   = 0;
	RuleStmt      *nothing  = NULL;
	RuleStmt      *rule     = NULL;
	RangeTblEntry *entry    = NULL;
	RangeTblEntry *viewrte  = NULL;
	char          *viewname = NULL;
	bool          replace_update   = FALSE;
	bool          replace_nothing  = FALSE;

	Assert( (tentries != NULL) 
			&& (rel != NULL)
			&& (var != NULL)
			&& (select != NULL) );

	baserel = get_reloid_from_select( select, &entry );
	
	if ( baserel > 0 ) 
	{
		
		Index newrti         = 0;
		Index baserti        = 0;
		Index oldrti         = 0;
		RangeTblRef *newref  = NULL;
		RangeTblRef *baseref = NULL;
		Query *gen_update    = NULL;
		Index rtindex        = 0;
		char option          = makeViewCheckOption(checkOption,
												   checkCascade);
		
		update              = makeNode( Query );
		update->commandType = CMD_UPDATE;
		
		/* 
		 * Catch the elog() from form_query(). We don't want
		 * to error out here, otherwise the whole CREATE VIEW
		 * command will fail.
		 */
		PG_TRY();
		{
			form_query( select, update, tentries, true );
		}
		PG_CATCH();
		{
			elog(DEBUG1, "could not create UPDATE update rule");
			return NULL;
		}
		PG_END_TRY();
		
		/*
		 * Our UPDATE rule needs range table references for
		 * the *NEW* and base relation.
		 *
		 * Note: the jointree in the UPDATE tree is already prepared, 
		 * we only had to fill the fromlist list.
		 */
		
		newrti = get_rtindex_for_rel( update->rtable, "*NEW*" );
		baserti = get_rtindex_for_rel( update->rtable, 
									   entry->eref->aliasname );
		oldrti = get_rtindex_for_rel( update->rtable,
									  "*OLD*" );
		
		newref = makeNode( RangeTblRef );
		baseref = makeNode( RangeTblRef );
		
		newref->rtindex = oldrti;
		baseref->rtindex = baserti;
		
		update->jointree->fromlist = list_make1( baseref );
		update->jointree->fromlist = lappend( update->jointree->fromlist,
											  newref );
		
		/*
		 * Create the WHERE condition qualification for the new rule
		 */
		form_where_for_updrule( select,
								&(update->jointree),
								rel,
								baserel,
								baserti,
								oldrti );
		
		/*
		 * We must reorder the columns in the targetlist to
		 * match the underlying table. We do this after
		 * calling form_where_for_updrule() because 
		 * build_update_target_list() relies on the
		 * original resdoms in the update tree.
		 */
		build_update_target_list( update,
								  select,
								  tentries,
								  baserel,
								  rel );
		
		/*
		 * Pass this generic tree to create_nothing_rule(), 
		 * it's sufficient to create an INSTEAD DO NOTHING
		 * rule for deletion.
		 * 
		 * This is required, because we can get conditional rules
		 * if the checkOption is specified. We only save this generic tree
		 * here, the work is done at the end of this function.
		 */
		
		gen_update = copyObject( update );
		gen_update->commandType = CMD_NOTHING;
		
		/*
		 * Get the relation name for the view
		 */
		rtindex = get_rtindex_for_rel( update->rtable, "*NEW*" );      
		Assert( rtindex > 0 );
		viewrte = rt_fetch( rtindex, update->rtable );
		viewname = get_rel_name( viewrte->relid );
		
		elog( DEBUG1, "viewname for relid %u is %s", viewrte->relid, viewname );
		
		/*
		 * Set ACL bit
		 */
		entry->requiredPerms |= ACL_UPDATE;
		
		/*
		 * UPDATE rule has to be replaced?
		 */
		if ( OidIsValid(hasRule( viewrte->relid, UPDATERULENAME ) ) ) 
		{	
			replace_update = TRUE;	
		} 
		else 
		{	
			replace_update = FALSE;	
		}
		
		/*
		 * NOTHING rule has to be replaced?
		 */
		if ( OidIsValid(hasRule( viewrte->relid, NOTHING_UPDATERULENAME ) ) ) 
		{	
			replace_nothing = TRUE;	
		} 
		else 
		{	
			replace_nothing = FALSE;
		}
		
		/*
		 * Ready to create the rule
		 */
		
		nothing = create_nothing_rule( gen_update,
									   var,
									   NOTHING_UPDATERULENAME,
									   CMD_UPDATE,
									   replace_nothing );
		
		rule = create_update_rule( update,
								   var,
								   replace_update );
		
		/*
		 * enter rule qualification creation
		 */
		if (checkOption)
			create_qual_for_rule( rule, select, tentries, rtindex, baserti );
		
		DefineQueryRewrite( nothing, NO_OPTION_IMPLICIT );
		DefineQueryRewrite( rule, option );
		
	}
	
	return update;  

}

/*------------------------------------------------------------------------------
 * Transforms the specified SELECT query tree into
 * an equivalent INSERT statement. Note:
 * The function assumes that the specified query tree
 * has passed the checkTree() function before.
 *------------------------------------------------------------------------------
 */
extern Query *
transform_select_to_insert( const Query       *select,
							const Relation    rel,
							const RangeVar    *var,
							ViewDefColumnList tentries,
							bool              checkOption,
							bool              checkCascade ) 
{

	Query         *insert   = NULL;
	Oid           baserel   = 0;
	RuleStmt      *nothing  = NULL;
	RuleStmt      *rule     = NULL;
	RangeTblEntry *entry    = NULL;
	RangeTblEntry *viewrte  = NULL;
	char          *viewname = NULL;
	bool          replace_insert  = FALSE;
	bool          replace_nothing = FALSE;

	Assert( (tentries != NULL) 
			&& (rel != NULL)
			&& (var != NULL)
			&& (select != NULL) );

	baserel = get_reloid_from_select( select, &entry );
	
	if ( baserel > 0 ) 
	{
		
		Query *gen_insert = NULL;
		Index rtindex     = 0;
		Index baserti     = 0;
		char option       = makeViewCheckOption(checkOption,
												checkCascade);
		
		/*
		 * Check if we can create the insert rule for this 
		 * view.
		 */
		if (!isInsertable(select, rel)) 
		{
			elog(DEBUG1, "Cannot create implicit insert rule for this view");
			return NULL;
		}
		
		insert              = makeNode( Query );
		insert->commandType = CMD_INSERT;
		
		/* 
		 * Catch the elog() from form_query(). We don't want
		 * to error out here, otherwise the whole CREATE VIEW
		 * command will fail.
		 */
		PG_TRY();
		{
			form_query( select, insert, tentries, true );
		}
		PG_CATCH();
		{
			elog(DEBUG1, "could not create INSERT update rule");
			return NULL;
		}
		PG_END_TRY();
		
		/*
		 * Pass this generic tree to create_nothing_rule(), 
		 * it's sufficient to create a INSTEAD DO NOTHING
		 * rule for deletion.
		 * 
		 * This is required, because we can get conditional rules
		 * if the checkOption is specified. We only save this generic tree
		 * here, the work is done at the end of this function.
		 */
		
		gen_insert = copyObject( insert );
		gen_insert->commandType = CMD_NOTHING;
		
		/*
		 * Get the relation name for the view
		 */
		baserti = get_rtindex_for_rel( insert->rtable,
									   entry->eref->aliasname );
		
		rtindex = get_rtindex_for_rel( insert->rtable, "*NEW*" );      
		Assert( rtindex > 0 );
		viewrte = rt_fetch( rtindex, insert->rtable );
		viewname = get_rel_name( viewrte->relid );
		
		elog( DEBUG1, "Viewname for relid %u is %s", viewrte->relid, viewname );
		
		/*
		 * We must reorder the columns in the targetlist to
		 * match the underlying table. We do this after
		 * calling form_where_for_updrule() because 
		 * build_update_target_list() relies on the
		 * original resdoms in the update tree.
		 */
		build_update_target_list( insert,
								  select,
								  tentries,
								  baserel,
								  rel );
		
		/*
		 * Set ACL bit
		 */
		entry->requiredPerms |= ACL_INSERT;
		
		/*
		 * INSERT rule has to be replaced?
		 */
		if ( OidIsValid(hasRule( viewrte->relid, INSERTRULENAME ) ) ) 
		{	
			replace_insert = TRUE;	
		} 
		else 
		{	
			replace_insert = FALSE;	
		}
		
		/*
		 * NOTHING rule has to be replaced?
		 */
		if ( OidIsValid(hasRule( viewrte->relid, NOTHING_INSERTRULENAME ) ) ) 
		{	
			replace_nothing = TRUE;	
		} 
		else 
		{
			replace_nothing = FALSE;
		}
		
		/*
		 * Ready to create the rule
		 */
		
		nothing = create_nothing_rule( gen_insert,
									   var,
									   NOTHING_INSERTRULENAME,
									   CMD_INSERT,
									   replace_nothing );
		
		rule = create_insert_rule( insert,
								   var,
								   replace_insert );
		
		/*
		 * enter rule qualification creation
		 */
		if (checkOption)
			create_qual_for_rule( rule, select, tentries, rtindex, baserti );
		
		DefineQueryRewrite( nothing, NO_OPTION_IMPLICIT );
		DefineQueryRewrite( rule, option );
		
	}
	
	return insert;

}

/*------------------------------------------------------------------------------
 * Checks the specified Query Tree for SQL92 compliance.
 * Returns FALSE if the specified SELECT query tree does not
 * pass the SQL92 requirements, otherwise TRUE is returned.
 *------------------------------------------------------------------------------
 */
extern bool
checkTree( const Query *query,
		   struct ViewBaseRelation *tree ) 
{

	bool     result     = true;
	List     *relations = NIL;
	ListCell *cell      = NULL;

	Assert( (query != NULL) && (tree != NULL) );

	if (!check_reltree( tree,
						&relations )) 
	{
		
		elog(DEBUG1, 
			 "possible JOIN/UNION in relations found in view definition");
		return false;

	}

	if ( query->commandType != CMD_SELECT ) 
	{
		elog(DEBUG1, "Specified OID is not a valid view _RETURN rule");
		return false;
	}

	if ( query->hasAggs == true ) 
	{
		elog(DEBUG1, "Aggregates violates SQL92 view update rules!");
		return false;
	}

	/*
	if ( query->hasSubLinks == true ) 
	{
		elog( WARNING, "Subqueries violates SQL92 view update rules!" );
		return false;
	}
	*/

	if ( list_length( query->groupClause ) >= 1 ) 
	{
		elog(DEBUG1, "GROUP BY violates SQL92 view update rules!");
		return false;
	}

	if ( list_length( query->distinctClause ) >= 1 ) 
	{
		elog(DEBUG1, "DISTINCT violates SQL92 view update rules!");
		return false;
	}

	if ( query->havingQual != NULL ) 
	{
		elog(DEBUG1, "HAVING violates SQL92 view update rules!");
		return false;
	}

	/*
	 * Test for number of involved relations. Since we assume
	 * to operate on a view definition SELECT query tree, we must
	 * count 3 rtable entries. Otherwise this seems not to be a 
	 * view based on a single relation.
	 */
	if ( list_length( query->rtable ) > 3 ) 
	{
		elog(DEBUG1, "View seems to have more than one base relation!");
		return false;
	}

	/*
	 * Any rtable entries involved?
	 */
	if ( list_length( query->rtable ) < 3 ) 
	{
		elog(DEBUG1, "No base relation detected!");
		return false;
	}

	/*
	 * Walk down the target list and look for nodes that aren't
	 * VARs. SQL92 doesn't allow functions, hostvariables or constant
	 * expressions in the col's target list.
	 * Also, check if any of the target list entries are
	 * indexed array expressions, which aren't supported.
	 */
	foreach(cell, query->targetList) 
	{

		Node *node = (Node *)lfirst(cell);

		if (IsA(node, TargetEntry)) 
		{

			if (!IsA(((TargetEntry *)node)->expr, Var) 
				&& !IsA(((TargetEntry *)node)->expr, ArrayRef))
			{					
				elog(DEBUG1, "view target list has invalid column entries");
				return false;				
			} else if (IsA(((TargetEntry *)node)->expr, ArrayRef))
			{
				ArrayRef *ref = (ArrayRef *)((TargetEntry *)node)->expr;

				if (ref->refupperindexpr != NIL)
				{
					elog(DEBUG1, "indexed array fields aren't updatable");
					return false;
				}
			}
			
		}
		
	}
	
	/*
	 * At least check if any RTE are valid. We have to look especially for
	 * table functions, which cannot be ever updatable.
	 */
	foreach(cell, query->rtable) 
	{

		RangeTblEntry *entry = (RangeTblEntry *)lfirst(cell);

		if (entry->rtekind != RTE_RELATION) 
		{
			elog(DEBUG1, "range table entry %s is not a relation!",
				 entry->eref->aliasname );
			return false;
		}

	}

	return result;

}

/*------------------------------------------------------------------------------
 * Traverse the specified relation tree.
 * The function stops at the base relations at the
 * leafs of the tree. If any of the relations has more than
 * one base relations, it is considered as a non-SQL92 updatable
 * view and FALSE is returned.
 *------------------------------------------------------------------------------
 */
extern bool
check_reltree( struct ViewBaseRelation *node,
			   List   **baserelations ) 
{

	Assert( node != NULL );
	Assert( baserelations != NULL );

	if (node->defs != NIL) 
	{

		ListCell *acell = NULL;
		foreach( acell, node->defs ) 
		{

			/*
			 * Walk down the tree
			 */
			
			struct ViewBaseRelation *relations =
				(struct ViewBaseRelation *)lfirst(acell);
			
			/*
			 * Set base relation definition list only, if not
			 * empty. Otherwise we always would return an empty list, then.
			 */
			
			if (relations->defs != NIL) 
			{
				*baserelations = relations->defs;

				if (list_length( relations->defs) > 1)
					return FALSE;

			}
			
		}

	}

	return TRUE;

}

/*------------------------------------------------------------------------------
 * Returns the OID from the specified SELECT query tree.
 * Returns with 0 if select holds no SELECT query tree, otherwise
 * the oid of the query relation.The function assumes
 * that the specified query tree was checked by a 
 * previous call to the checkTree() function.
 *------------------------------------------------------------------------------
 */
extern Oid
get_reloid_from_select( const Query   *select,
						RangeTblEntry **rel_entry ) 
{

	Oid      result = 0;
	ListCell *cell  = NULL;	

	/*
	 * Set RTE to null, to indicate nothing is found.
	 */
	*rel_entry = NULL;

	/*
	 * Check specified query tree. Return on error immediately.
	 */

	if ( select == NULL)
		return result;

	if ( select->commandType != CMD_SELECT ) 
		return result;

	/*
	 * We loop through the RTEs to get information about
	 * all involved relations. Note, that we return the
	 * first OID we found in the list, which is *NOT* a
	 * *NEW* or *OLD* range entry. 
	 */

	if ( list_length(select->rtable) > 0 ) 
	{

		foreach( cell, select->rtable ) 
		{

			RangeTblEntry *entry  = (RangeTblEntry *)lfirst( cell );
      
			if ( entry != NULL ) 
			{

				elog( DEBUG1, "Extracted range table entry for %u", 
					  entry->relid );

				/*
				 * Look for *OLD* and *NEW* entries. If found, we can be sure, this is
				 * a view rule. We went over to the next RTEs, since we need the involved
				 * base relation.
				 */

				if ( entry->eref ) 
				{

					if ( strcmp( entry->eref->aliasname, "*OLD*" ) == 0 ) 
						continue;

					if ( strcmp( entry->eref->aliasname, "*NEW*" ) == 0 ) 
						continue;

					if ( strlen( entry->eref->aliasname ) > 0 ) {

						elog( DEBUG1, "Return base relation OID %u", 
							  entry->relid );
						result = entry->relid;
						*rel_entry = copyObject( entry );

					}

				}

			} 
			else
				elog(ERROR, 
					 "RTE null pointer in get_reloid_from_select()");

		}

	}

	return result;

}

/*------------------------------------------------------------------------------
 * Required for update rule CHECK OPTION
 *------------------------------------------------------------------------------
 */
PG_FUNCTION_INFO_V1(pg_view_update_error);

Datum
pg_view_update_error( PG_FUNCTION_ARGS ) 
{
  
  bool qual = PG_GETARG_BOOL(0);

  if (PG_ARGISNULL(0)) 
  {
	  /*
	   * A view's CHECK OPTION should never evaluate to NULL,
	   * so we treat a NULL evaluated expression tree as FALSE.
	   */
	  qual = FALSE;
  }

  if (qual == FALSE)
	  ereport( ERROR,
			   (errcode( ERRCODE_WITH_CHECK_OPTION_VIOLATION ), 
				errmsg("View update commands violates rule condition!" )));
  
  PG_RETURN_BOOL(true);

}

